"""Module to create engines for data processing."""

# standard library
from abc import ABC, abstractmethod
from dataclasses import dataclass
import os
from typing import Dict, List, Union

# third party libraries
import duckdb
from duckdb import DuckDBPyRelation
import polars as pl
import pyarrow as pa

# local libraries
from datagrunt.core.csvcomponents import CSVDelimiter, CSVColumnNameNormalizer
from datagrunt.core.queries import DuckDBQueries

@dataclass
class EngineProperties:
    """Base properties for CSV operations."""
    filepath: str
    dataframe_sample_rows: int = 20
    csv_export_filename: str = 'output.csv'
    excel_export_filename: str = 'output.xlsx'
    json_export_filename: str = 'output.json'
    json_newline_export_filename: str = 'output.jsonl'
    parquet_export_filename: str = 'output.parquet'
    valid_engines: tuple = ('duckdb', 'polars')
    value_error_message: str = """Reader engine '{engine}' is not 'duckdb' or 'polars'. Pass either 'duckdb' or 'polars' as valid engine params."""
    missing_file_message: str = """File '{filepath}'. No such file or directory."""

class CSVBaseReaderEngine(ABC):
    """Abstract base class defining the interface for reader engines."""

    def __init__(self, filepath):
        """Initialize the CSVReader class.

        Args:
            filepath (str): Path to the file to read.
        """
        self.filepath = filepath
        self.queries = DuckDBQueries(self.filepath)
        self.db_table = DuckDBQueries(self.filepath).database_table_name
        self.delimiter = CSVDelimiter(self.filepath).delimiter
        if not os.path.exists(self.filepath):
            raise FileNotFoundError

    @abstractmethod
    def get_sample(self, normalize_columns: bool = False) -> None:
        """Return a sample of the data.

        Args:
            normalize_columns (bool): Whether to normalize column names.
        """
        pass

    @abstractmethod
    def to_dataframe(self, normalize_columns: bool = False) -> pl.DataFrame:
        """Convert data to a dataframe.

        Args:
            normalize_columns (bool): Whether to normalize column names.
        """
        pass

    @abstractmethod
    def to_arrow_table(self, normalize_columns: bool = False) -> pa.Table:
        """Convert data to a PyArrow table.

        Args:
            normalize_columns (bool): Whether to normalize column names.
        """
        pass

    @abstractmethod
    def to_dicts(self, normalize_columns: bool = False) -> List[Dict]:
        """Convert data to a list of dictionaries.

        Args:
            normalize_columns (bool): Whether to normalize column names.
        """
        pass

    @abstractmethod
    def query_data(self, sql_query: str, normalize_columns: bool = False) -> Union[DuckDBPyRelation, pl.DataFrame]:
        """Query the data using SQL.

        Args:
            sql_query (str): SQL query to execute.
            normalize_columns (optional, bool): Whether to normalize column names.
        """
        pass

class CSVBaseWriterEngine(ABC):
    """Abstract base class defining the interface for writer engines."""

    def __init__(self, filepath):
        """
        Initialize the CSV Writer DuckDB Engine class.

        Args:
            filepath (str): Path to the file to write.
        """
        self.filepath = filepath
        self.queries = DuckDBQueries(self.filepath)
        self.db_table = DuckDBQueries(self.filepath).database_table_name
        if not os.path.exists(self.filepath):
            raise FileNotFoundError

    @abstractmethod
    def write_csv(self, export_filename, normalize_columns=False):
        """Write data to CSV format."""
        pass

    @abstractmethod
    def write_excel(self, export_filename, normalize_columns=False):
        """Write data to Excel format.

        Args:
            export_filename (str): Path to the file to write.
            normalize_columns (optional, bool): Whether to normalize column names.
        """
        pass

    @abstractmethod
    def write_json(self, export_filename, normalize_columns=False):
        """Write data to JSON format."""
        pass

    @abstractmethod
    def write_json_newline_delimited(self, export_filename, normalize_columns=False):
        """Write data to JSON Lines format.

        Args:
            export_filename (str): Path to the file to write.
            normalize_columns (optional, bool): Whether to normalize column names.
        """
        pass

    @abstractmethod
    def write_parquet(self, export_filename, normalize_columns=False):
        """Write data to Parquet format.

        Args:
            export_filename (str): Path to the file to write.
            normalize_columns (optional, bool): Whether to normalize column names.
        """
        pass

class CSVReaderDuckDBEngine(CSVBaseReaderEngine):
    """Class to read CSV files and convert CSV files powered by DuckDB."""

    def get_sample(self, normalize_columns=False):
        """Return a sample of the CSV file.

        Args:
            normalize_columns (optional, bool): Whether to normalize column names.
        """
        self.queries.create_table(normalize_columns).show()

    def to_dataframe(self, normalize_columns=False):
        """Converts CSV to a Polars dataframe.

        Args:
            normalize_columns (optional, bool): Whether to normalize column names.

        Returns:
            A Polars dataframe.
        """
        return self.queries.create_table(normalize_columns).pl()

    def to_arrow_table(self, normalize_columns=False):
        """Converts CSV to a PyArrow table.

        Args:
            normalize_columns (optional, bool): Whether to normalize column names.

        Returns:
            A PyArrow table.
        """
        return self.queries.create_table(normalize_columns).arrow()

    def to_dicts(self, normalize_columns=False):
        """Converts CSV to a list of Python dictionaries.

        Args:
            normalize_columns (optional, bool): Whether to normalize column names.

        Returns:
            A list of dictionaries.
        """
        return self.to_dataframe(normalize_columns).to_dicts()

    def _normalize_relation_columns(self, relation):
        """Applies column name normalization to a DuckDBPyRelation.
        Args:
            relation (duckdb.DuckDBPyRelation): The relation to normalize.
        Returns:
            duckdb.DuckDBPyRelation: The normalized relation.
        """
        current_columns = relation.columns
        column_normalizer = CSVColumnNameNormalizer(self.filepath)
        projections = []
        for col in current_columns:
            normalized_name = column_normalizer.columns_to_normalized_mapping.get(col, col)
            projections.append(f'"{col}" AS "{normalized_name}"')
        return relation.project(", ".join(projections))

    def query_data(self, sql_query, normalize_columns=False):
        """Queries as CSV file after importing into DuckDB.

        Args:
            sql_query (str): Query to run against DuckDB.
            normalize_columns (optional, bool): Whether to normalize column names.

        Returns:
            A DuckDB DuckDBPyRelation with the query results.

        Example if DuckDB Engine:
            dg = CSVReader('myfile.csv')
            query = f"SELECT col1, col2 FROM {dg.db_table}"
            dg.query_csv_data(query)
        """
        # Ensure the base table is created with original column names
        # so the user's query can reference them.
        self.queries.create_table(normalize_columns=False)

        # Execute the user's query
        result_relation = duckdb.sql(sql_query)

        if normalize_columns:
            result_relation = self._normalize_relation_columns(result_relation)

        return result_relation

class CSVReaderPolarsEngine(CSVBaseReaderEngine):
    """Class to read CSV files and convert CSV files powered by Polars."""

    def _create_dataframe(self, normalize_columns=False):
        """Normalizes the column names of the dataframe.

        Args:
            normalize_columns (optional, bool): Whether to normalize column names.

        Returns:
            A Polars dataframe.
        """
        df = pl.read_csv(self.filepath,
                         separator=self.delimiter,
                         truncate_ragged_lines=True,
                         infer_schema=False
                        )
        if normalize_columns:
            df = df.rename(CSVColumnNameNormalizer(self.filepath).columns_to_normalized_mapping)
        return df

    def _create_dataframe_sample(self, normalize_columns=False):
        """Create a sample of the CSV file as a Polars dataframe.

        Args:
            normalize_columns (optional, bool): Whether to normalize column names.

        Returns:
            A Polars dataframe.
        """
        df = pl.read_csv(self.filepath,
                         separator=self.delimiter,
                         truncate_ragged_lines=True,
                         infer_schema=False,
                         n_rows=EngineProperties.dataframe_sample_rows
                        )
        if normalize_columns:
            df = df.rename(CSVColumnNameNormalizer(self.filepath).columns_to_normalized_mapping)
        return df

    def get_sample(self, normalize_columns=False):
        """Return a sample of the CSV file.

        Args:
            normalize_columns (optional, bool): Whether to normalize column names.

        Returns:
            A Polars dataframe.
        """
        df = self._create_dataframe_sample(normalize_columns)
        print(df)

    def to_dataframe(self, normalize_columns=False):
        """Converts CSV to a Polars dataframe.

        Args:
            normalize_columns (optional, bool): Whether to normalize column names.

        Returns:
            A Polars dataframe.
        """
        return self._create_dataframe(normalize_columns)

    def to_arrow_table(self, normalize_columns=False):
        """Converts CSV to a PyArrow table.

        Args:
            normalize_columns (optional, bool): Whether to normalize column names.

        Returns:
            A PyArrow table.
        """
        df = self._create_dataframe(normalize_columns).to_arrow()
        return df

    def to_dicts(self, normalize_columns=False):
        """Converts CSV to a list of Python dictionaries.

        Args:
            normalize_columns (optional, bool): Whether to normalize column names.

        Returns:
            A list of dictionaries.
        """
        dicts = self._create_dataframe(normalize_columns).to_dicts()
        return dicts

    def query_data(self, sql_query, normalize_columns=False):
        """Queries as CSV file after importing into DuckDB.

        Args:
            sql_query (str): Query to run against DuckDB.
            normalize_columns (optional, bool): Whether to normalize column names.

        Returns:
            A DuckDB DuckDBPyRelation with the query results.

        Example if DuckDB Engine:
            dg = CSVReader('myfile.csv')
            query = "SELECT col1, col2 FROM {dg.db_table}" # f string assumed
            dg.query_csv_data(query)
        """
        return self.queries.sql_query_to_dataframe(sql_query, normalize_columns)

class CSVWriterDuckDBEngine(CSVBaseWriterEngine):
    """Class to convert CSV files to various other supported file types powered by DuckDB."""

    def write_csv(self, export_filename=None, normalize_columns=False):
        """Query to export a DuckDB table to a CSV file.

            Args:
                export_filename str: The name of the output file.
                normalize_columns bool: Whether to normalize column names.
            """
        filename = self.queries.set_export_filename(EngineProperties.csv_export_filename, export_filename)
        self.queries.create_table(normalize_columns)
        duckdb.sql(self.queries.export_csv_query(filename))

    def write_excel(self, export_filename=None, normalize_columns=False):
        """Query to export a DuckDB table to an Excel file.

        Args:
            export_filename (optional, str): The name of the output file.
            normalize_columns (optional, bool): Whether to normalize column names.
        """
        filename = self.queries.set_export_filename(EngineProperties.excel_export_filename, export_filename)
        self.queries.create_table(normalize_columns)
        duckdb.sql(self.queries.export_excel_query(filename))

    def write_json(self, export_filename=None, normalize_columns=False):
        """Query to export a DuckDB table to a JSON file.

        Args:
            export_filename (optional, str): The name of the output file.
            normalize_columns (optional, bool): Whether to normalize column names.
        """
        filename = self.queries.set_export_filename(EngineProperties.json_export_filename, export_filename)
        self.queries.create_table(normalize_columns)
        duckdb.sql(self.queries.export_json_query(filename))

    def write_json_newline_delimited(self, export_filename=None, normalize_columns=False):
        """Query to export a DuckDB table to a JSON newline delimited file.

        Args:
            export_filename (optional, str): The name of the output file.
            normalize_columns (optional, bool): Whether to normalize column names.
        """
        filename = self.queries.set_export_filename(EngineProperties.json_newline_export_filename, export_filename)
        self.queries.create_table(normalize_columns)
        duckdb.sql(self.queries.export_json_newline_delimited_query(filename))

    def write_parquet(self, export_filename=None, normalize_columns=False):
        """Query to export a DuckDB table to a Parquet file.

        Args:
            export_filename (optional, str): The name of the output file.
            normalize_columns (optional, bool): Whether to normalize column names.
        """
        filename = self.queries.set_export_filename(EngineProperties.parquet_export_filename, export_filename)
        self.queries.create_table(normalize_columns)
        duckdb.sql(self.queries.export_parquet_query(filename))

class CSVWriterPolarsEngine(CSVBaseWriterEngine):
    """Class to write CSVs to other file formats powered by Polars."""

    def write_csv(self, export_filename=None, normalize_columns=False):
        """Export a Polars dataframe to a CSV file.

        Args:
            export_filename (optional, str): The name of the output file.
            normalize_columns (optional, bool): Whether to normalize column names.
        """
        filename = self.queries.set_export_filename(EngineProperties.csv_export_filename, export_filename)
        df = CSVReaderPolarsEngine(self.filepath).to_dataframe(normalize_columns)
        df.write_csv(filename)

    def write_excel(self, export_filename=None, normalize_columns=False):
        """Export a Polars dataframe to an Excel file.

        Args:
            export_filename (optional, str): The name of the output file.
            normalize_columns (optional, bool): Whether to normalize column names.
        """
        filename = self.queries.set_export_filename(EngineProperties.excel_export_filename, export_filename)
        df = CSVReaderPolarsEngine(self.filepath).to_dataframe(normalize_columns)
        df.write_excel(filename)

    def write_json(self, export_filename=None, normalize_columns=False):
        """Export a Polars dataframe to a JSON file.

        Args:
            export_filename (optional, str): The name of the output file.
            normalize_columns (optional, bool): Whether to normalize column names.
        """
        filename = self.queries.set_export_filename(EngineProperties.json_export_filename, export_filename)
        df = CSVReaderPolarsEngine(self.filepath).to_dataframe(normalize_columns)
        df.write_json(filename)

    def write_json_newline_delimited(self, export_filename=None, normalize_columns=False):
        """Export a Polars dataframe to a JSON newline delimited file.

        Args:
            export_filename (optional, str): The name of the output file.
            normalize_columns (optional, bool): Whether to normalize column names.
        """
        filename = self.queries.set_export_filename(EngineProperties.json_newline_export_filename, export_filename)
        df = CSVReaderPolarsEngine(self.filepath).to_dataframe(normalize_columns)
        df.write_ndjson(filename)

    def write_parquet(self, export_filename=None, normalize_columns=False):
        """Export a Polars dataframe to a Parquet file.

        Args:
            export_filename (optional, str): The name of the output file.
            normalize_columns (optional, bool): Whether to normalize column names.
        """
        filename = self.queries.set_export_filename(EngineProperties.parquet_export_filename, export_filename)
        df = CSVReaderPolarsEngine(self.filepath).to_dataframe(normalize_columns)
        df.write_parquet(filename)



# Import key classes, functions, or submodules that should be available at the package level
from datagrunt.csvfiles.csvreader import CSVReader
from datagrunt.csvfiles.csvwriter import CSVWriter

# You can define __all__ to specify what gets imported with "from package import *"
__all__ = ['CSVReader', 'CSVWriter']



"""Module for writing CSV files and converting to different file formats."""

# standard library

# third party libraries

# local libraries
from datagrunt.core import CSVComponents
from datagrunt.core import DuckDBQueries
from datagrunt.core import CSVEngineFactory

class CSVWriter(CSVComponents):
    """Class to unify the interface for converting CSV files to various other supported file types."""

    def __init__(self, filepath, engine='duckdb'):
        """Initialize the CSV Writer class.

        Args:
            filepath (str): Path to the file to write.
            engine (str, default 'duckdb'): Determines which writer engine class to instantiate.
        """
        super().__init__(filepath)
        self.db_table = DuckDBQueries(self.filepath).database_table_name
        self.engine = engine.lower().replace(' ', '')

    def _create_writer(self):
        """Create a reader object."""
        return CSVEngineFactory(self.filepath, self.engine).create_writer()

    def write_csv(self, out_filename=None, normalize_columns=False):
        """Query to export a DuckDB table to a CSV file.

            Args:
                out_filename str: The name of the output file.
                normalize_columns optional, bool: Whether to normalize column names.
            """
        return self._create_writer().write_csv(out_filename, normalize_columns)

    def write_excel(self, out_filename=None, normalize_columns=False):
        """Query to export a DuckDB table to an Excel file.

        Args:
            out_filename str: The name of the output file.
            normalize_columns optional, bool: Whether to normalize column names.
        """
        return self._create_writer().write_excel(out_filename, normalize_columns)

    def write_json(self, out_filename=None, normalize_columns=False):
        """Query to export a DuckDB table to a JSON file.

        Args:
            out_filename str: The name of the output file.
            normalize_columns optional, bool: Whether to normalize column names.
        """
        return self._create_writer().write_json(out_filename, normalize_columns)

    def write_json_newline_delimited(self, out_filename=None, normalize_columns=False):
        """Query to export a DuckDB table to a JSON newline delimited file.

        Args:
            out_filename str: The name of the output file.
            normalize_columns optional, bool: Whether to normalize column names.
        """
        return self._create_writer().write_json_newline_delimited(out_filename, normalize_columns)

    def write_parquet(self, out_filename=None, normalize_columns=False):
        """Query to export a DuckDB table to a Parquet file.

        Args:
            out_filename str: The name of the output file.
            normalize_columns optional, bool: Whether to normalize column names.
        """
        return self._create_writer().write_parquet(out_filename, normalize_columns)



"""Module for CSV components."""

# standard library
from collections import Counter, OrderedDict
import csv
from functools import lru_cache
import re

# third party libraries
import polars as pl

# local libraries
from datagrunt.core.fileproperties import FileProperties

class CSVDelimiter:
    """Class to infer and derive the CSV delimiter."""

    DELIMITER_REGEX_PATTERN = r'[^0-9a-zA-Z_ "-]'
    DEFAULT_DELIMITER = ','
    DEFAULT_TAB_DELIMITER = '\t'

    def __init__(self, filepath):
        super().__init__()
        """Initialize the CSVDelimiter class.

        Args:
            filepath (str): The path to the CSV file.
        """
        self.file_properties = FileProperties(filepath)
        self.first_row = CSVRows(filepath).first_row
        self.delimiter = self.infer_csv_file_delimiter()
        self.delimiter_byte_string = self.delimiter.encode()

    def _get_most_common_non_alpha_numeric_character_from_string(self):
        """Get the most common non-alpha-numeric character from a given string.

        Returns:
            str: The most common non-alpha-numeric character from the string.
        """
        columns_no_spaces = self.first_row.replace(' ', '')
        regex = re.compile(self.DELIMITER_REGEX_PATTERN)
        counts = Counter(char for char in regex.findall(columns_no_spaces))
        most_common = counts.most_common()
        return most_common

    def infer_csv_file_delimiter(self):
        """Infer the delimiter of a CSV file.

        Returns:
            str: The delimiter of the CSV file.
        """
        delimiter_candidates = self._get_most_common_non_alpha_numeric_character_from_string()

        if self.file_properties.is_empty or self.file_properties.is_blank:
            delimiter = self.DEFAULT_DELIMITER
        elif self.file_properties.is_tsv:
            delimiter = self.DEFAULT_TAB_DELIMITER
        elif len(delimiter_candidates) == 0:
            delimiter = ' '
        else:
            delimiter = delimiter_candidates[0][0]
        return delimiter

class CSVDialect:
    """Class for inferring the CSV dialect."""

    CSV_SNIFF_SAMPLE_ROWS = 5
    QUOTING_MAP = {
        0: 'no quoting',
        1: 'quote all',
        2: 'quote minimal',
        3: 'quote non-numeric'
    }

    def __init__(self, filepath):
        """Initialize the CSVDialect object.

        Args:
            filepath (str): The path to the CSV file.
        """
        self.filepath = filepath
        self.dialect = self._get_csv_dialect()

    def _get_csv_dialect(self):
        """Get the CSV dialect from the file.

        Returns:
            csv.Dialect: The CSV dialect inferred from the file.
        """
        with open(self.filepath, 'r', encoding=FileProperties(self.filepath).DEFAULT_ENCODING) as csvfile:
            dialect = csv.Sniffer().sniff(csvfile.read(self.CSV_SNIFF_SAMPLE_ROWS))
            csvfile.seek(0)  # Reset file pointer to the beginning
        return dialect

    @property
    def quotechar(self):
        """The character used to quote fields in the CSV file."""
        return self.dialect.quotechar

    @property
    def escapechar(self):
        """The character used to escape characters in the CSV file."""
        return self.dialect.escapechar

    @property
    def doublequote(self):
        """Whether double quotes are used to escape quotes in the CSV file."""
        return self.dialect.doublequote

    @property
    def newline_delimiter(self):
        """The newline delimiter used in the CSV file."""
        return self.dialect.lineterminator

    @property
    def skipinitialspace(self):
        """Whether spaces are skipped at the beginning of fields in the CSV file."""
        return self.dialect.skipinitialspace

    @property
    def quoting(self):
        """The quoting style used in the CSV file."""
        return self.QUOTING_MAP.get(self.dialect.quoting)

class CSVRows:
    """Class for parsing CSV rows."""

    def __init__(self, filepath):
        """Initialize the CSVRows object.

        Args:
            filepath: The path to the CSV file.
        """
        self.filepath = filepath
        self.first_row = self._get_first_row_from_file()

    def _get_first_row_from_file(self):
        """Reads and returns the first line of a file.

        Returns:
            The first line of the file, stripped of leading/trailing whitespace,
            or None if the file is empty.
        """
        with open(self.filepath, 'r', encoding=FileProperties(self.filepath).DEFAULT_ENCODING) as csv_file:
            first_line = csv_file.readline().strip()
        return first_line

    @property
    @lru_cache()
    def row_count_with_header(self):
        """Return the number of lines in the CSV file including the header."""
        with open(self.filepath, 'rb') as csv_file:
            return sum(1 for _ in csv_file)

    @property
    def row_count_without_header(self):
        """Return the number of lines in the CSV file excluding the header."""
        return self.row_count_with_header - 1

class CSVColumns:
    """Class for parsing CSV columns."""

    def __init__(self, filepath):
        """Initialize the CSVColumns class."""
        self.filepath = filepath
        self.columns = self._get_columns()

    def _get_columns(self):
        """Return the columns."""
        df = pl.read_csv(self.filepath,
                         separator=CSVDelimiter(self.filepath).delimiter,
                         truncate_ragged_lines=True,
                         infer_schema=False,
                         n_rows=5
                        )
        return df.columns

    @property
    def columns_string(self):
        """Return a string representation of the columns."""
        return ', '.join(self.columns)

    @property
    def columns_byte_string(self):
        """Return a byte string representation of the columns."""
        return ', '.join(self.columns).encode()

    @property
    def columns_count(self):
        """Return the number of columns."""
        return len(self.columns)

class CSVColumnNameNormalizer:
    """Class to normalize CSV columns names."""

    SPECIAL_CHARS_PATTERN = re.compile(r'[^a-z0-9]+')
    MULTI_UNDERSCORE_PATTERN = re.compile(r'_+')

    def __init__(self, filepath):
        """Initialize the CSVColumnNameNormalizer with a filepath."""
        self.filepath = filepath
        self.columns_normalized = self._normalize_column_names(CSVColumns(filepath).columns)

    def _normalize_single_column_name(self, column_name):
        """Normalize a single column name by converting to lowercase, replacing spaces and special
        characters with underscores, and removing extra underscores.

        Replace special characters and spaces with underscore
        Remove leading and trailing underscores
        Replace multiple underscores with single underscore
        Add a leading underscore if the name starts with a digit

        Args:
            column_name (str): The column name to normalize

        Returns:
            str: The normalized column name
        """
        name = column_name.lower()
        name = self.SPECIAL_CHARS_PATTERN.sub('_', name)
        name = name.strip('_')
        name = self.MULTI_UNDERSCORE_PATTERN.sub('_', name)
        return f'_{name}' if name and name[0].isdigit() else name

    def _make_unique_column_names(self, columns_list):
        """
        Make unique column names by appending a number to duplicate names.

        Args:
            columns (list): List of column names to make unique

        Returns:
            list: List of unique column names
        """
        name_count = {}
        unique_names = []

        for name in columns_list:
            if name in name_count:
                name_count[name] += 1
                unique_names.append(f"{name}_{name_count[name]}")
            else:
                name_count[name] = 0
                unique_names.append(name)

        return unique_names

    def _normalize_column_names(self, columns):
        """
        Normalize column names by converting to lowercase, replacing spaces and special
        characters with underscores, and removing extra underscores.

        Args:
            columns (list): List of column names to normalize

        Returns:
            list: List of normalized column names
        """
        normalized_columns = [self._normalize_single_column_name(col) for col in columns]
        return self._make_unique_column_names(normalized_columns)

    @property
    def columns_normalized_string(self):
        """Return a list representation of the normalized columns."""
        return ', '.join(self.columns_normalized)

    @property
    def columns_normalized_byte_string(self):
        """Return a byte string representation of the normalized columns."""
        return ', '.join(self.columns_normalized).encode()

    @property
    def columns_to_normalized_mapping(self):
        """Return the mapping of original column names to normalized column names."""
        return dict(OrderedDict(zip(CSVColumns(self.filepath).columns, self.columns_normalized)))

class CSVComponents(FileProperties):
    def __init__(self, filepath):
        """Initialize the CSVComponents object.

        Args:
            filepath (str): Path to the CSV file.
        """
        super().__init__(filepath)
        self.delimiter = CSVDelimiter(filepath).delimiter

    @property
    def quotechar(self):
        """Return the quote character used in the CSV file."""
        return CSVDialect(self.filepath).quotechar

    @property
    def escapechar(self):
        """Return the escape character used in the CSV file."""
        return CSVDialect(self.filepath).escapechar

    @property
    def doublequote(self):
        """Return the double quote character used in the CSV file."""
        return CSVDialect(self.filepath).doublequote

    @property
    def newline_delimiter(self):
        """Return the newline delimiter used in the CSV file."""
        return CSVDialect(self.filepath).newline_delimiter

    @property
    def skipinitialspace(self):
        """Return the skipinitialspace flag used in the CSV file."""
        return CSVDialect(self.filepath).skipinitialspace

    @property
    def quoting(self):
        """Return the quoting flag used in the CSV file."""
        return CSVDialect(self.filepath).quoting

    @property
    def row_count_with_header(self):
        """Return the number of rows in the CSV file including the header row."""
        return CSVRows(self.filepath).row_count_with_header

    @property
    def row_count_without_header(self):
        """Return the number of rows in the CSV file excluding the header row."""
        return CSVRows(self.filepath).row_count_without_header

    @property
    def columns(self):
        """Return the columns of the CSV file."""
        return CSVColumns(self.filepath).columns

    @property
    def columns_string(self):
        """Return the columns of the CSV file as a string."""
        return CSVColumns(self.filepath).columns_string

    @property
    def columns_byte_string(self):
        """Return the columns of the CSV file as a byte string."""
        return CSVColumns(self.filepath).columns_byte_string

    @property
    def columns_count(self):
        """Count the columns of the CSV file."""
        return CSVColumns(self.filepath).columns_count

    @property
    def columns_normalized(self):
        """Normalize the columns of the CSV file."""
        return CSVColumnNameNormalizer(self.filepath).columns_normalized

    @property
    def columns_normalized_string(self):
        """Normalize the columns of the CSV file."""
        return CSVColumnNameNormalizer(self.filepath).columns_normalized_string

    @property
    def columns_normalized_byte_string(self):
        """Normalize the columns of the CSV file."""
        return CSVColumnNameNormalizer(self.filepath).columns_normalized_byte_string

    @property
    def columns_to_normalized_mapping(self):
        """Normalize the columns of the CSV file."""
        return CSVColumnNameNormalizer(self.filepath).columns_to_normalized_mapping



"""Module for deriving and evaluating file properties."""

# standard library
import os
from pathlib import Path

class FileExtensions:
    """Class for getting file extensions."""
    def __init__(self, filepath):
        """Initialize the FileExtensions object.

        Args:
            filepath (str): Path to the file.
        """
        self.filepath = filepath

    @property
    def extension(self):
        """Get the file extensions."""
        return Path(self.filepath).suffixes

    @property
    def csv_extensions(self):
        """Define CSV extensions."""
        return ['csv']

    @property
    def tsv_extensions(self):
        """Define TSV extensions."""
        return ['tsv']

    @property
    def excel_extensions(self):
        """Define Excel extensions."""
        return [
            'xlsx',
            'xlsm',
            'xlsb',
            'xltx',
            'xltm',
            'xls',
            'xlt'
        ]

    @property
    def tabular_extensions(self):
        """Define tabular extensions."""
        return list(set(self.csv_extensions + self.tsv_extensions + self.excel_extensions))

    @property
    def apache_extensions(self):
        """Define Apache extensions."""
        return ['parquet', 'avro']

    @property
    def semi_structured_extensions(self):
        """Define semi-structured extensions."""
        return list(set(['json', 'jsonl']))

    @property
    def standard_extensions(self):
        """Define standard (non proprietary) extensions."""
        return list(set(self.csv_extensions +
                        self.tsv_extensions +
                        self.apache_extensions +
                        self.semi_structured_extensions))

    @property
    def structured_extensions(self):
        """Define structured extensions."""
        return list(set(self.csv_extensions +
                        self.tsv_extensions +
                        self.excel_extensions +
                        self.apache_extensions +
                        self.tabular_extensions))

    @property
    def proprietary_extensions(self):
        """Define proprietary extensions."""
        return list(set(self.excel_extensions))

class FileStatistics:
    """Class for getting file statistics."""

    FILE_SIZE_DIVISOR = 1_000
    EXCEL_ROW_LIMIT = 1_048_576
    FILE_SIZE_ROUND_FACTOR = 5
    LARGE_FILE_FACTOR = 1.0 # size in GB

    def __init__(self, filepath):
        """Initialize the FileStatistics object.

        Args:
            filepath (str): Path to the file.
        """
        self.filepath = filepath
        self.size_in_bytes = os.path.getsize(self.filepath)
        self.size_in_kb = round((self.size_in_bytes / self.FILE_SIZE_DIVISOR), self.FILE_SIZE_ROUND_FACTOR)
        self.size_in_mb = round((self.size_in_kb / self.FILE_SIZE_DIVISOR), self.FILE_SIZE_ROUND_FACTOR)
        self.size_in_gb = round((self.size_in_mb / self.FILE_SIZE_DIVISOR), self.FILE_SIZE_ROUND_FACTOR)
        self.size_in_tb = round((self.size_in_gb / self.FILE_SIZE_DIVISOR), self.FILE_SIZE_ROUND_FACTOR)

    @property
    def modified_time(self):
        """Get the file modified time."""
        return os.path.getmtime(self.filepath)

    @property
    def is_large(self):
        """Check if the file is at least one gigabyte or larger."""
        return self.size_in_gb >= self.LARGE_FILE_FACTOR

class BlankFile:
    """Class for checking if a file is blank."""

    FILE_SIZE_MB_FACTOR = 10.0

    def __init__(self, filepath):
        """Initialize the BlankFile object.

        Args:
            filepath (str): Path to the file.
        """
        self.filepath = filepath

    @property
    def is_blank(self):
        """Check if the file is blank. Blank files contain only whitespace."""

        filestats = FileStatistics(self.filepath)

        # Very low probability of being blank if file is 10MB or larger in size
        if filestats.size_in_mb >= self.FILE_SIZE_MB_FACTOR:
            return False
        with open(self.filepath, 'r') as f:
            content = f.read().strip()
            if not content:
                return True
        return False

class EmptyFile:
    """Class for checking if a file is empty."""
    def __init__(self, filepath):
        """Initialize the EmptyFile object.

        Args:
            filepath (str): Path to the file.
        """
        self.filepath = filepath

    @property
    def is_empty(self):
        """Check if the file is empty."""
        return FileStatistics(self.filepath).size_in_bytes == 0

class FileProperties:
    """Base class for file objects."""

    DEFAULT_ENCODING = 'utf-8'

    def __init__(self, filepath):
        """
        Initialize the FileBase class.

        Args:
            filepath (str): Path to the file.
        """
        self.filepath = filepath
        self.filename = Path(filepath).name
        self.extension = Path(filepath).suffix
        self.extension_string = self.extension.replace('.', '')
        self.size_in_bytes = FileStatistics(self.filepath).size_in_bytes
        self.size_in_kb = FileStatistics(self.filepath).size_in_kb
        self.size_in_mb = FileStatistics(self.filepath).size_in_mb
        self.size_in_gb = FileStatistics(self.filepath).size_in_gb
        self.size_in_tb = FileStatistics(self.filepath).size_in_tb

    @property
    def is_structured(self):
        """Check if the file is structured."""
        return self.extension_string.lower() in FileExtensions(self.filepath).structured_extensions

    @property
    def is_semi_structured(self):
        """Check if the file is semi-structured."""
        return self.extension_string.lower() in FileExtensions(self.filepath).semi_structured_extensions

    @property
    def is_unstructured(self):
        """Check if the file is unstructured."""
        return self.extension_string.lower() not in FileExtensions(self.filepath).standard_extensions and \
               self.extension_string.lower() not in FileExtensions(self.filepath).semi_structured_extensions

    @property
    def is_standard(self):
        """Check if the file is standard."""
        return self.extension_string.lower() in FileExtensions(self.filepath).standard_extensions

    @property
    def is_proprietary(self):
        """Check if the file is proprietary."""
        return self.extension_string.lower() in FileExtensions(self.filepath).proprietary_extensions

    @property
    def is_csv(self):
        """Check if the file is a CSV file."""
        return self.extension_string.lower() in FileExtensions(self.filepath).csv_extensions

    @property
    def is_excel(self):
        """Check if the file is an Excel file."""
        return self.extension_string.lower() in FileExtensions(self.filepath).excel_extensions

    @property
    def is_apache(self):
        """Check if the file is an Apache formatted file."""
        return self.extension_string.lower() in FileExtensions(self.filepath).apache_extensions

    @property
    def is_empty(self):
        """Check if the file is empty. Empty files have a size of 0 bytes."""
        return EmptyFile(self.filepath).is_empty

    @property
    def is_blank(self):
        """Check if the file is blank. Blank files contain only whitespace."""
        return BlankFile(self.filepath).is_blank

    @property
    def is_large(self):
        """Check if the file is greater than or equal to 1 GB."""
        return FileStatistics(self.filepath).is_large

    @property
    def is_tabular(self):
        """Check if the file is tabular."""
        return self.extension_string.lower() in FileExtensions(self.filepath).tabular_extensions

    @property
    def is_tsv(self):
        """Check if the file is tabular."""
        return self.extension_string.lower() in FileExtensions(self.filepath).tsv_extensions


"""Module to store database queries and query strings."""

# standard library

# third party libraries
import duckdb

# local libraries
from datagrunt.core.databases import DuckDBDatabase
from datagrunt.core.csvcomponents import CSVDelimiter, CSVColumns, CSVColumnNameNormalizer

class DuckDBQueries:
    """Class to store DuckDB database queries and query strings."""

    def __init__(self, filepath):
        """
        Initialize the DuckDBQueries class.

        Args:
            filepath (str): Path to the file.
        """
        self.filepath = filepath
        self.delimiter = CSVDelimiter(filepath).delimiter
        self.database_table_name = DuckDBDatabase(filepath).database_table_name

    def set_export_filename(self, default_filename, export_filename=None):
        """Evaluate if a filename is passed in and if not, return default filename."""
        # This method doesn't really fit the class but it's the only place it's relevant.
        if export_filename:
            filename = export_filename
        else:
            filename = default_filename
        return filename

    def import_csv_query(self):
        """Query to import a CSV file into a DuckDB table.

        Returns:
            str: The query to import the CSV file.
        """
        return f"""
            CREATE OR REPLACE TABLE {self.database_table_name} AS
            SELECT *
            FROM read_csv('{self.filepath}',
                            auto_detect=true,
                            delim='{self.delimiter}',
                            header=true,
                            null_padding=true,
                            all_varchar=True,
                            strict_mode=false);
            """

    def import_csv_query_normalize_columns(self):
        """Query to import a CSV file into a DuckDB table and normalize column names.

        Returns:
            str: The query to import the CSV file and normalize column names.
        """
        return f"""
            CREATE OR REPLACE TABLE {self.database_table_name} AS
            SELECT *
            FROM read_csv('{self.filepath}',
                            auto_detect=true,
                            delim='{self.delimiter}',
                            header=true,
                            null_padding=true,
                            all_varchar=True,
                            strict_mode=false,
                            normalize_names=true);
            """

    def select_from_duckdb_table(self):
        """Query to select from a DuckDB table."""
        return f"SELECT * FROM {self.database_table_name}"

    def export_csv_query(self, default_filename, export_filename=None):
        """Query to export a DuckDB table to a CSV file.

        Args:
            default_filename (str): The default name of the output file.
            export_filename (str, optional): The name of the output file.

        Returns:
            str: The SQL query to export the table to a CSV file.
        """
        filename = self.set_export_filename(default_filename, export_filename)
        return f"COPY {self.database_table_name} TO '{filename}' (HEADER, DELIMITER ',');"

    def export_excel_query(self, default_filename, export_filename=None):
        """Query to export a DuckDB table to an Excel file.

        Args:
            default_filename (str): The default name of the output file.
            export_filename (str, optional): The name of the output file.

        Returns:
            str: The SQL query to export the table to an Excel file.
        """
        filename = self.set_export_filename(default_filename, export_filename)
        return f"""
            INSTALL spatial;
            LOAD spatial;
            COPY (SELECT * FROM {self.database_table_name})
            TO '{filename}'(FORMAT GDAL, DRIVER 'xlsx')
        """

    def export_json_query(self, default_filename, export_filename=None):
        """Query to export a DuckDB table to a JSON file.

        Args:
            default_filename (str): The default name of the output file.
            export_filename (str, optional): The name of the output file.

        Returns:
            str: The SQL query to export the table to a JSON file.
        """
        filename = self.set_export_filename(default_filename, export_filename)
        return f"COPY (SELECT * FROM {self.database_table_name}) TO '{filename}' (ARRAY true) "

    def export_json_newline_delimited_query(self, default_filename, export_filename=None):
        """Query to export a DuckDB table to a JSON file with newline delimited.

        Args:
            default_filename (str): The default name of the output file.
            export_filename (str, optional): The name of the output file.

        Returns:
            str: The SQL query to export the table to a JSON file with newline delimited.
        """
        filename = self.set_export_filename(default_filename, export_filename)
        return f"COPY (SELECT * FROM {self.database_table_name}) TO '{filename}'"

    def export_parquet_query(self, default_filename, export_filename=None):
        """Query to export a DuckDB table to a Parquet file.

        Args:
            default_filename (str): The default name of the output file.
            export_filename (str, optional): The name of the output file.

        Returns:
            str: The SQL query to export the table to a Parquet file.
        """
        filename = self.set_export_filename(default_filename, export_filename)
        return f"COPY (SELECT * FROM {self.database_table_name}) TO '{filename}'(FORMAT PARQUET)"

    def update_and_normalize_column_names(self):
        """Query to update column names in a DuckDB table.

        DuckDB has a built-in function to normalize column names. However,
        the format of the column names from the native DuckDB function often
        differs from the class CSVColumnNameNormalizer. Because other types
        of engines throughout the ecosystem may have different conventions,
        this method uses the CSVColumnNameNormalizer class to ensure
        consistent naming conventions across different processing engines.
        """
        duckdb.sql(self.import_csv_query())
        for old_name, new_name in zip(CSVColumns(self.filepath).columns,
                                      CSVColumnNameNormalizer(self.filepath).columns_normalized
                                      ):
            sql_string = f"ALTER TABLE {self.database_table_name} RENAME COLUMN \"{old_name}\" TO \"{new_name}\""
            duckdb.sql(sql_string)

    def create_table(self, normalize_columns=False):
        """Create a DuckDB table from the CSV file.

        Args:
            normalize_columns (bool): Whether to normalize column names.
        """
        if normalize_columns:
            self.update_and_normalize_column_names()
        else:
            duckdb.sql(self.import_csv_query())
        return duckdb.sql(self.select_from_duckdb_table()).execute()

    def _normalize_dataframe_columns(self, dataframe):
        """Applies column name normalization to a Polars DataFrame.
        Args:
            dataframe (polars.DataFrame): The DataFrame to normalize.
        Returns:
            polars.DataFrame: The normalized DataFrame.
        """
        column_normalizer = CSVColumnNameNormalizer(self.filepath)
        normalized_mapping = {
            col: column_normalizer.columns_to_normalized_mapping.get(col, col)
            for col in dataframe.columns
        }
        return dataframe.rename(normalized_mapping)

    def sql_query_to_dataframe(self, sql_query, normalize_columns=False):
        """Query to convert a SQL query to a Polars DataFrame.

        Args:
            sql_query (str): The SQL query to execute.
            normalize_columns (optional, bool): Whether to normalize column names.

        Returns:
            polars.DataFrame: The resulting DataFrame.
        """
        # Ensure the table is created with original column names for querying
        duckdb.sql(self.import_csv_query())

        # Execute the user's query
        result_df = duckdb.sql(sql_query).pl()

        if normalize_columns:
            result_df = self._normalize_dataframe_columns(result_df)

        return result_df



"""Module for interfacing with databases."""

# standard library
import os
from pathlib import Path
import re

# third party libraries
import duckdb

class DuckDBDatabase:
    """Class to configure local database for file processing.
       Utilizes duckdb as the processing engine.
    """
    DEFAULT_ENCODING = 'utf-8'
    DEFAULT_THREAD_COUNT = 16

    def __init__(self, filepath):
        """
        Initialize the FileDatabase class.

        Args:
            filepath (str): Path to the file.
        """
        self.filepath = filepath
        self.database_filename = self._set_database_filename()
        self.database_table_name = self._set_database_table_name()
        self.database_connection = self._set_database_connection()

    def __del__(self):
        """
        Close the database connection and delete .db files after use.
        """
        self.database_connection.close()
        if os.path.exists(self.database_filename):
            os.remove(self.database_filename)

    def _format_filename_string(self):
        """Remove all non alphanumeric characters from filename."""
        return re.sub(r'[^a-zA-Z0-9]', '', Path(self.filepath).stem)

    def _set_database_filename(self):
        """Return name of duckdb file created at runtime."""
        return f'{self._format_filename_string()}.db'

    def _set_database_table_name(self):
        """Return name of duckdb import table created during file import."""
        return f'{self._format_filename_string()}'

    def _set_database_connection(self, threads=DEFAULT_THREAD_COUNT):
        """Establish a connection with duckdb.

        Args:
            threads (int): Number of threads to use for duckdb.
        """
        return duckdb.connect(self.database_filename, config = {'threads': threads})



"""Module for reading CSV files and converting to different in memory python objects."""

# standard library

# third party libraries
import polars as pl

# local libraries
from datagrunt.core import CSVComponents
from datagrunt.core import DuckDBQueries
from datagrunt.core import CSVEngineFactory

class CSVReader(CSVComponents):
    """Class to unify the interface for reading CSV files."""

    def __init__(self, filepath, engine='polars'):
        """Initialize the CSV Reader class.

        Args:
            filepath (str): Path to the file to read.
            engine (str, default 'polars'): Determines which reader engine class to instantiate.
        """
        super().__init__(filepath)
        self.db_table = DuckDBQueries(self.filepath).database_table_name
        self.engine = engine.lower().replace(' ', '')

    def _return_empty_file_object(self, object):
        """Return an empty file object."""
        return object

    def _create_reader(self):
        """Create a reader object."""
        return CSVEngineFactory(self.filepath, self.engine).create_reader()

    def get_sample(self, normalize_columns=False):
        """Return a sample of the CSV file."""
        self._create_reader().get_sample(normalize_columns)

    def to_dataframe(self, normalize_columns=False):
        """Converts CSV to a Polars dataframe.

        Args:
            normalize_columns (bool): Whether to normalize column names.

        Returns:
            A Polars dataframe.
        """
        if self.is_empty or self.is_blank:
            return self._return_empty_file_object(pl.DataFrame())
        return self._create_reader().to_dataframe(normalize_columns)

    def to_arrow_table(self, normalize_columns=False):
        """Converts CSV to a PyArrow table.

        Args:
            normalize_columns (bool): Whether to normalize column names.

        Returns:
            A PyArrow table.
        """
        if self.is_empty or self.is_blank:
            return self._return_empty_file_object(pl.DataFrame().to_arrow())
        return self._create_reader().to_arrow_table(normalize_columns)

    def to_dicts(self, normalize_columns=False):
        """Converts CSV to a list of dictionaries.

        Args:
            normalize_columns (bool): Whether to normalize column names.

        Returns:
            A list of dictionaries.
        """
        if self.is_empty or self.is_blank:
            return self._return_empty_file_object(list())
        return self._create_reader().to_dicts(normalize_columns)

    def query_data(self, sql_query, normalize_columns=False):
        """Queries as CSV file after importing into DuckDB.

        Args:
            sql_query (str): Query to run against DuckDB.
            normalize_columns (optional, bool): Whether to normalize column names.

        Returns:
            A DuckDB DuckDBPyRelation with the query results.

        Example if DuckDB Engine:
            dg = CSVReader('myfile.csv')
            query = "SELECT col1, col2 FROM {dg.db_table}" # f string assumed
            dg.query_csv_data(query)

        If you set normalize_columns=True, the column names will be normalized to lowercase
        and spaces will be replaced with underscores, and you must reference the new column names
        in your query.
        """
        if self.is_empty or self.is_blank:
            return self._return_empty_file_object(list())
        return self._create_reader().query_data(sql_query, normalize_columns)



"""
Datagrunt

A Python library designed to simplify the way you work with CSV files.

This module provides inferred CSV delimiters and helper methods for reading and writing CSV files.

Example:
    A simple example of how to use the main functionality of your package:

    from datagrunt.csvfile import CSVReader

    csv_file = 'electric_vehicle_population_data.csv'
    engine = 'duckdb'

    dg = CSVReader(csv_file, engine=engine)

    dg.get_sample()

Attributes:
    __version__: A string representing the version of this module.
    __author__: The name of the package author.
    __license__: The license under which the package is released.
"""

__version__ = "1.0.16"
__author__ = "Martin Graham"
__license__ = "MIT"

# Import key classes, functions, or submodules that should be available at the package level
from datagrunt.csvfiles import CSVReader, CSVWriter


# You can define __all__ to specify what gets imported with "from package import *"
__all__ = ['CSVReader', 'CSVWriter']

# Optionally, you can include a logger for your package
import logging

logging.getLogger(__name__).addHandler(logging.NullHandler())



"""Factory module for creating factory instances."""

# standard library
import os

# third party libraries

# local libraries
from datagrunt.core.engines import EngineProperties
from datagrunt.core.engines import (
    CSVReaderDuckDBEngine,
    CSVReaderPolarsEngine,
    CSVWriterDuckDBEngine,
    CSVWriterPolarsEngine
)

from datagrunt.core.queries import DuckDBQueries

class CSVEngineFactory:
    """Factory class for creating CSV reader and writer engine instances."""

    READER_ENGINES = {
                'duckdb': CSVReaderDuckDBEngine,
                'polars': CSVReaderPolarsEngine,
            }

    WRITER_ENGINES = {
                'duckdb': CSVWriterDuckDBEngine,
                'polars': CSVWriterPolarsEngine,
            }

    def __init__(self, filepath, engine):
        """
        Initialize the Engine Factory class.

        Args:
            filepath (str): Path to the file to read.
            engine (str): type of engine to create by the factory.
        """
        self.filepath = filepath
        self.engine = engine.lower().replace(' ', '')
        self.db_table = DuckDBQueries(self.filepath).database_table_name
        if not os.path.exists(self.filepath):
            raise FileNotFoundError
        if self.engine not in EngineProperties.valid_engines:
            raise ValueError(EngineProperties.value_error_message.format(engine=self.engine))

    def create_reader(self):
        """Create a reader engine instance.

        Args:
            filepath (str): Path to the input file
            engine (str): Engine type ('duckdb' or 'polars')

        Returns:
            An instance of BaseReaderEngine
        """
        engine_class = self.READER_ENGINES.get(self.engine)
        if engine_class:
            return engine_class(self.filepath)
        else:
            raise ValueError(f"Unsupported reader engine: {self.engine}")

    def create_writer(self):
        """Create a writer engine instance.

        Args:
            filepath (str): Path to the input file
            engine (str): Engine type ('duckdb' or 'polars')

        Returns:
            An instance of BaseWriterEngine
        """
        engine_class = self.WRITER_ENGINES.get(self.engine)
        if engine_class:
            return engine_class(self.filepath)
        else:
            raise ValueError(f"Unsupported reader engine: {self.engine}")



from datagrunt.core.databases import DuckDBDatabase
from datagrunt.core.queries import DuckDBQueries
from datagrunt.core.engines import EngineProperties
from datagrunt.core.factories import CSVEngineFactory
from datagrunt.core.fileproperties import FileProperties

from datagrunt.core.csvcomponents import (
    CSVDelimiter,
    CSVDialect,
    CSVColumns,
    CSVColumnNameNormalizer,
    CSVComponents,
    CSVRows,
)

__all__ = ['DuckDBDatabase',
           'DuckDBQueries',
           'CSVEngineFactory',
           'EngineProperties',
           'FileProperties',
           'CSVDelimiter',
           'CSVDialect',
           'CSVColumns',
           'CSVColumnNameNormalizer',
           'CSVComponents',
           'CSVRows',
]
